# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0

#
"""All necessary constants for the package."""

from typing import Final


SPEED_OF_LIGHT: Final[float] = 299792458.0
r"""Speed of light in vacuum in :math:`\text{m}/\text{s}`"""
MSUN_SI: Final[float] = 1.9884099021470415e30
"""Solar mass in kg."""
SECONDS_PER_YEAR: Final[float] = 365.25 * 24 * 60 * 60
"""Seconds in 1 year."""
