# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


#
"""
.. admonition::
    🚧 Page under construction 🚧


    Documentation for this module is under construction. You can check the source code for more information or
    run :code:`--help` on the following CLIs:

    - :code:`sage_one_powerlaw_one_peak`
    - :code:`ppd_one_powerlaw_one_peak`
"""
