# Copyright 2023 The GWKokab Authors
# SPDX-License-Identifier: Apache-2.0


#
"""Literals used in Kokab."""

from typing import Final


LOG_REF_PRIOR_NAME: Final[str] = "log_prior"
"""Helper variable to store the column name of log reference prior."""


INFERENCE_DIRECTORY: Final[str] = "inference"
"""Name of the directory to store inference results."""
